//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     nints.h
// Purpose:    Set of interpretations for critical default negations.
// Last Change:12.03.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2002 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	NINTS_INCLUDED
#define NINTS_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef BOOL_INCLUDED
#include "bool.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef NINT_INCLUDED
#include "nint.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// NINTS_MAGIC: Magic number (identifies objects of nints_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	NINTS_MAGIC =	0x4E495300L;	// 'NIS\0'
#endif

//-----------------------------------------------------------------------------
// NINTS_FREE_MAGIC: Magic number for nints_c objects on free list.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	NINTS_FREE_MAGIC = 0x5F495300L;	// '_IS\0'
#endif

//-----------------------------------------------------------------------------
// NINTS_HASHSIZE: Size of the hash table.
//-----------------------------------------------------------------------------

static const int	NINTS_HASHSIZE = 1001; // Should be a prime number.

//=============================================================================
// Class for Elements in the Current Set of Default Negation Interpretations.
//=============================================================================

class nints_c {

//-----------------------------------------------------------------------------
// Static Member Functions (Class Methods):
//-----------------------------------------------------------------------------

	public:

	// clear: Reinitialize to empty set (reuse memory).
	static void clear(void);

	// insert: Insert default negation interpretation if not already there.
	static void insert(nint_t neg_int);

	// print_all: Print all currently existing interpretations.
	static void print_all(void);

	// get_first: Get first interpretation (to loop through the set).
	static inline nints_c* get_first(void)
		{ CHECK(Initialized, "nints_c::get_first: Not initialized");
		  return(First);
		}

//-----------------------------------------------------------------------------
// Member Functions (Object Methods):
//-----------------------------------------------------------------------------

	public:

	// Get pointer to next interpretation:
	inline nints_c* get_next(void) const
		{ CHECK(valid(), "nints_c::get_next: Bad object");
		  return(Next); }

	// Get default negation in this set element:
	inline nint_t get_neg_int(void) const
		{ CHECK(valid(), "nints_c::get_neg_int: Bad object");
		  return(NegInt); }

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "NINTS_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Private Member Functions (Auxillary Object Methods):
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	inline nints_c(void)
		{
#ifdef VER_DEBUG
		  Magic = NINTS_FREE_MAGIC;
#endif
		}

	// Destructor:
	inline ~nints_c(void)
		{
#ifdef VER_DEBUG
		  Magic = 0;
#endif
		}

	// Initialization of new or reused object:
	inline void hash_enter(nints_c**hashtab_entry, nint_t neg_int)
		{ CHECK(Magic == NINTS_FREE_MAGIC,
			"nints_c::hash_enter: object not new");
		  CHECK(*hashtab_entry == 0 || (*hashtab_entry)->valid(),
			"nints_c::hash_enter: hash table entry invalid");
		  CHECK(nint_c::valid(neg_int),
			"nints_c::hash_enter: neg_int is invalid");
		  NegInt = neg_int;
		  Next = 0;
		  HashLink = *hashtab_entry;
		  *(hashtab_entry) = this;
		  HashTabEntry = hashtab_entry;
#ifdef VER_DEBUG
		  Magic = NINTS_MAGIC;
#endif
		}

	// Replacement for destructor (put into freelist):
	inline void	free(void)
		{ CHECK(this != 0, "nints_c::free: object is NIL");
		  CHECK(Magic == NINTS_MAGIC, "nints_c::free: invalid object");
		  *(HashTabEntry) = 0;
			// Unlinks the entire list, but is anyway called only
			// in clear().
		  Next = Free;
		  Free = this;
#ifdef VER_DEBUG
		  Magic = NINTS_FREE_MAGIC;
#endif
		}

	// Set the Next pointer:
	inline void set_next(nints_c *next)
		{ CHECK(valid(), "nints_c::set_next: Bad object");
		  CHECK(next == 0 || next->valid(),
		  	  "nints_c::set_next: next is invalid");
		  Next = next; }

	// Get the next element of the element chain for this hash table entry:
	inline nints_c* get_hashlink(void) const
		{ CHECK(valid(), "nints_c::get_hashlink: Bad object");
		  return(HashLink); }

	// Return the next element of the free list:
	inline nints_c *get_next_free(void) const
		{ CHECK(this != 0, "nints_c::get_next_free: object is NIL");
		  CHECK(Magic == NINTS_FREE_MAGIC,
			"nints_c::get_next_free: Bad object");
		  return(Next); }

//-----------------------------------------------------------------------------
// Static Data Members (Class Attributes):
//-----------------------------------------------------------------------------

	private:

	// Was the entire hash table set to NIL?
	static bool_t Initialized;

	// The hash table:
	static nints_c* HashTab[NINTS_HASHSIZE];

	// First element in the list of all default negation interpretations:
	static nints_c* First;

	// Last element in the list of all default negation interpretations:
	static nints_c* Last;

	// Deleted objects (can be reused):
	static nints_c* Free;

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Default negation interpretation (key in this tree node):
	nint_t	NegInt;

	// Link to next default negation:
	nints_c* Next;

	// Link in the hash table chain:
	nints_c* HashLink;

	// Link back to hash table entry (for cleaning hash table in clear()):
	nints_c** HashTabEntry;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef nints_c *nints_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

